//####################################################
//#     This software released to the public domain  #
//#     without restrictions of any kind.            #
//####################################################

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <string.h>
#include <process.h>
#include <ctype.h>
#include <i86.h>
#include <sys/sched.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/qnxterm.h>
#include <sys/proxy.h>
#include <sys/kernel.h>
#include <sys/dev.h>
#include <sys/timers.h>
#include <sys/wait.h>
#include <sys/sys_msg.h>
#include <sys/ser_msg.h>
#include <sys/name.h>
#include "eqn.h"

#define STDERR  0 
#define false	0
#define true    !false

#define CHNLMAX	16
#define PAT_ORG 32
#define PAT_LEN 64
#define WBUFSIZE 512    /* must be multiple of PAT_LEN */
#define RBUFSIZE 4096   /* must be multiple of PAT_LEN */

struct chnl_struct
{
  char com_link[CHNLMAX];
  int com_fd;
  pid_t in_tid;
  unsigned char in_stk[1024 + RBUFSIZE];
  unsigned long in_count;
  unsigned char in_nextchar;
  unsigned long in_errs;
  pid_t out_tid;
  unsigned char out_stk[1024 + WBUFSIZE];
  unsigned long out_count;
  unsigned long out_rate;
};  /* chnl_struct */

/* define global space for the individual channel structures */
struct chnl_struct chnl[CHNLMAX];

/* note that globals are available to threads */
int chnlfirst;
int chnlcnt = 1;
int loopon = false;
int loopavail = true;
int baud = 115200;
int rx_check = false;
int flowon = true;
char *orgport;
char portgrp[128];
pid_t drvr_pid;
char *p;
char *q;

/* forward prototypes */
int eqn_loop( int fd, int mode, pid_t drvr_pid );
void eqnwrite( void *arg );
void eqnread( void *arg );
void eqnbench_usage();
void rx_ck( unsigned char *buf, int len, struct chnl_struct *chnl );
void terminate( int signal );

/******************************************************************************

	MAIN (eqnbench)	

	This program exercises 1 to 16 serial ports simultaneously at
	one common baudrate.  Data is handled "raw", 8-bits, no parity.

	main spawns up to 16 writers and 16 readers.  The readers use 
	QNX's dev_read() to manage input.  

	Statistics are reported on the screen managed by main.  The
	data for the statistics is supplied by the reader and writer
	threads.  QNX "term_" functions are utilized to handle the 
	screen.
	
******************************************************************************/

main( int argc, char *argv[] )
{
int ii, jj;
char ss[128];
char tt[128];
struct termios termios;
struct timespec start, stop; 
double elapsed;
int avg;
int pct;
int ideal;

  if ( argc < 2 )
  {
    eqnbench_usage();
    exit(-1);
  }

  while ( (ii = getopt( argc,
                        argv, 
                        "?lvfb:c::" )) != -1 )
    switch ( ii ) 
    {
      case '?':
        eqnbench_usage();
        exit(-1);
      case 'l':
        loopon = true;
        break;
      case 'v':
        rx_check = true;
        break;
      case 'f':
        flowon = false;
        break;
      case 'b':
        baud = atoi(optarg);
        break;
      case 'c':
        chnlcnt = atoi(optarg);
        break;
    }
    orgport = argv[optind];
    if ( !strlen( orgport ) )
      exit( -2 );

#if STDERR
  fprintf( stderr, "%d %d %d %d %d %s\n", loopon, rx_check, flowon, baud, chnlcnt, orgport );
#endif

  /* get pid of eqn driver  */
  drvr_pid = qnx_name_locate( getnid(), EQN_PID_NAME, 0, NULL );
  if ( drvr_pid == -1 )
  {
    loopavail = false;
    loopon = false;
  }

#if STDERR
  fprintf( stderr, "drvr_pid: %d\n", drvr_pid );
#endif

  strcpy( portgrp, orgport );
  p = &portgrp[ strlen(portgrp) - 1 ];
  q = &orgport[ strlen(orgport) - 1 ];
  while ( (q > orgport) && isdigit(*q) )
  {
    q--;
    p--;
  }
  if ( q == orgport )
    exit( -3 );
  chnlfirst = atoi(p+1);
  if ( chnlfirst < 1 ) 
    exit( -4 );
  *(++p) = 0;  /* portgrp now set */
#if STDERR
  fprintf( stderr, "%s %s %d\n", orgport, portgrp, chnlfirst );
#endif

  if ( term_load() == -1 )
    exit( -5 );
  term_video_on();
  term_clear( 0 );

  if ( chnlcnt > CHNLMAX )
    chnlcnt = CHNLMAX;
  if ( chnlcnt > 1 )
    sprintf( ss, "%s[%d-%d]", portgrp, chnlfirst, chnlfirst + chnlcnt - 1 );
  else
    sprintf( ss, "%s", orgport );
  term_printf( 0, 60, 0, "%s", ss );

  if ( loopon )
    sprintf( ss, "loopback, " );
  else
    ss[0] = 0;
  if ( rx_check )
    strcat( ss, "verify" );
  else
    strcat( ss, "no verify" );
  term_printf( 1, 60, 0, "%s", ss );

  /* setup chnl array */
  for ( ii = 0, jj = chnlfirst; ii < chnlcnt; ii++, jj++ )
  {
    strcpy( ss, portgrp );
    sprintf( tt, "%d", jj );
    strcat( ss, tt );
    chnl[ii].in_nextchar = PAT_ORG;
    strcpy( chnl[ii].com_link, ss );
#if STDERR
    fprintf( stderr, "%s\n", chnl[ii].com_link);
#endif
    if ( (chnl[ii].com_fd = open( chnl[ii].com_link, O_RDWR )) == -1 )
    {
      perror( "eqnbench open failed" );
      exit(-6);
    }
    tcgetattr( chnl[ii].com_fd, &termios );
    termios.c_oflag &= ~OPOST;
    termios.c_iflag &= ~( BRKINT | ICRNL | IGNCR | INLCR | INPCK 
                           | ISTRIP | PARMRK | IXON | IXOFF );
    if ( !flowon )
      termios.c_iflag |= ( IGNBRK | IGNPAR );
    else
      termios.c_iflag |= ( IGNBRK | IGNPAR | IXON | IXOFF );

    termios.c_cflag &= ~( CSIZE | CSTOPB | PARENB | PARODD 
                            | PARSTK | IHFLOW | OHFLOW ); 
    termios.c_cflag |= ( CLOCAL | CREAD | CS8 | HUPCL );
    termios.c_lflag &= ~( ECHO | ECHOE | ECHOK | ECHONL | ICANON
                            | ISIG | NOFLSH | TOSTOP | IEXTEN );
    termios.c_cc[VTIME] = 0;
    termios.c_cc[VMIN]  = 0;
    termios.c_ispeed = baud;
    termios.c_ospeed = baud;
    tcsetattr( chnl[ii].com_fd, TCSADRAIN, &termios );
    tcflush( chnl[ii].com_fd, TCIOFLUSH );
    if ( ( chnl[ii].in_tid = tfork( chnl[ii].in_stk,
                                    sizeof( chnl[ii].in_stk),
                                    eqnread,
                                    &chnl[ii],
                                    0 ) ) < 0 )
    {
      /* could not fork */
      perror( "Fork failed...reader" );
      exit(-7);
    }

    if ( ( chnl[ii].out_tid = tfork( chnl[ii].out_stk,
                                     sizeof( chnl[ii].out_stk),
                                     eqnwrite,
                                     &chnl[ii],
                                     0 ) ) < 0 )
    {
      /* could not fork */
      perror( "Fork failed...writer" );
      exit(-8);
    }
  } /* init loop */

  for ( ii = 0; ii < chnlcnt; ii++ )
      close( chnl[ii].com_fd );

  getclock(TIMEOFDAY, &start);

  /* initialize signals */
  for( ii = _SIGMIN; ii <= _SIGMAX; ++ii )
    signal( ii, SIG_IGN );
  signal( SIGTERM, &terminate );
  signal( SIGQUIT, &terminate );
  signal( SIGINT, &terminate );
                                         
  ideal = baud / 10;

  /* parent continues execution here */
  term_printf( 0, 0, 0, "  Tot Xmit       Tot Rcv        Tot Errs       Rate" );

#if STDERR
  fprintf( stderr, "leader is pid: %d\n", getpid() );
#endif

  /* lower priority */
  setprio( 0, getprio( 0 ) - 1 );

  while ( true )
  {
    sleep(1);
    getclock(TIMEOFDAY, &stop);
    elapsed = (stop.tv_sec - start.tv_sec) + (stop.tv_nsec - start.tv_nsec)
                    / (float)1000000000L;
    avg = 0;
    for ( ii = 0; ii < chnlcnt; ii++ )
    {
      /* compute rates */
      chnl[ii].out_rate = chnl[ii].out_count / elapsed;
      avg += chnl[ii].out_rate;

      /* update screen */
      term_printf( ii + 2, 0, 0, "%10.1d    %10.1d    %10.1d    %10.1d", 
                   chnl[ii].out_count, 
                   chnl[ii].in_count,
                   chnl[ii].in_errs,
                   chnl[ii].out_rate );
    }
    avg /= chnlcnt;
    term_printf( 0, 51, 0, "%6.1d", avg );
    pct = 100 * avg / ideal;
    term_printf( 1, 54, 0, "%3.1d%s", pct, "%" );
    term_home();
    term_flush();
  }
}  /* main */

/******************************************************************************

	EQNREAD

	This is a "thread" routine.  One copy will be run for every
	channel opened.

	Its sole purpose is to read any data available on its assigned
	channel, to optionally check for errors on the received data
	 and to update its statistics counts.

******************************************************************************/

void eqnread( void *arg )
{
struct chnl_struct *chnlx;
unsigned char rstuff[RBUFSIZE];
pid_t proxy, pid;
int rdcnt;
int armed;
int ii, this_chnl;

  chnlx = (struct chnl_struct *) arg;

  /* get a proxy id */
  proxy = qnx_proxy_attach( 0, 0, 0, -1 );

  this_chnl = chnlx - chnl;
  for ( ii = 0; ii < chnlcnt; ii++ )
  {
    if ( ii != this_chnl )
      close( chnl[ii].com_fd );
  }
  close( 0 );
  close( 1 );
#if !STDERR
  close( 2 );
#endif

  /* lower priority */
  setprio( 0, getprio( 0 ) - 1 );

  /* issue a zero byte read to trigger a message in the main loop
       - starts normal processing */
  rdcnt = dev_read( chnlx->com_fd, rstuff, 0, 1, 0, 0, proxy, 0 );
#if STDERR
  fprintf( stderr, "reader #%d\n", chnlx - chnl ); 
#endif

  while ( true )
  {
    pid = Receive( 0, rstuff, RBUFSIZE );
    if ( pid == proxy )
    {
      armed = 0;
      while ( !armed )
      {
        rdcnt = dev_read( chnlx->com_fd, rstuff, RBUFSIZE, 
                          1, 10, 10, 
                          proxy, &armed );
        chnlx->in_count += rdcnt;
        if ( rx_check )
          rx_ck( rstuff, rdcnt, chnlx );
      }
    }
  }
}  /* eqnread */

/******************************************************************************

	EQNWRITE

	This is a "thread" routine.  One copy will be run for every
	channel opened.

	Its sole purpose is to write data on its assigned channel 
	and to update its statistics count.

******************************************************************************/

void eqnwrite( void *arg )
{
struct chnl_struct *chnlx;
int ii, this_chnl;
unsigned char wstuff[WBUFSIZE];
int wstuff_len;

  chnlx = (struct chnl_struct *) arg;

#if STDERR
  fprintf( stderr, "write #%d\n", chnlx - chnl ); 
#endif

  /* set loopback */
  if ( loopavail )
    if ( loopon )
      eqn_loop( chnlx->com_fd, 1, drvr_pid );
    else
      eqn_loop( chnlx->com_fd, 0, drvr_pid );

  wstuff_len = sizeof( wstuff );
  {
  char *p;

    p = wstuff;
    for ( ii = 0; ii < WBUFSIZE; ii++ )
      *p++ = ii % PAT_LEN + PAT_ORG;
  }
                         
  this_chnl = chnlx - chnl;
  for ( ii = 0; ii < chnlcnt; ii++ )
  {
    if ( ii != this_chnl )
      close( chnl[ii].com_fd );
  }
  close( 0 );
  close( 1 );
  close( 2 );

  /* lower priority */
  setprio( 0, getprio( 0 ) - 1 );
  while ( true )
  {
    write( chnlx->com_fd, wstuff, wstuff_len );
    chnlx->out_count += wstuff_len;
  }
}  /* eqnwrite */

/*****************************************************************************

	RX_CK

	Verify that "len" bytes at "buf" are the expected sequence.
	Keep track of the last character received so that the next time
	rx_ck is called, it will know which byte to expect.

*****************************************************************************/

void rx_ck( unsigned char *buf, int len, struct chnl_struct *achnl )
{
unsigned char c;
unsigned char *p;
int ii;

  c = achnl->in_nextchar;
  p = buf;
  ii = 0;
  while ( ii++ < len )
  {
    if ( *p++ != c++ )
    {
      achnl->in_errs++;
#if STDERR
      if ( achnl == chnl )
        fprintf( stderr, "found %d, expected %d, index %d, len %d\n", *(p-1), c-1, ii-1, len );
#endif
      break;
    }
    else
      if ( c >= PAT_ORG + PAT_LEN )
        c = PAT_ORG;
  }
  achnl->in_nextchar = buf[len - 1] + 1;
  /* if junk received, set to a valid value */
  if ( achnl->in_nextchar < PAT_ORG )
    achnl->in_nextchar = PAT_ORG;
  if ( achnl->in_nextchar >= PAT_ORG + PAT_LEN )
    achnl->in_nextchar = PAT_ORG;
}  /* rx_ck */

/*****************************************************************************

	TERMINATE

	Signal action routine responsible for clean termination of each
	channel.

*****************************************************************************/

void terminate( int signal )
{
int ii;
struct termios termios;

  for ( ii = 0; ii < chnlcnt; ii++ )
  {
    /* wait for this channel's reader and writer to die */
    waitpid( chnl[ii].in_tid, NULL, 0 );
    waitpid( chnl[ii].out_tid, NULL, 0 );
    /* clear loopback if necessary */
    if ( loopavail && loopon )
      eqn_loop( chnl[ii].com_fd, 0, drvr_pid );
    /* flush channel */
    termios.c_iflag &= ~( IXON | IXOFF );
    tcsetattr( chnl[ii].com_fd, TCSANOW, &termios );
    tcflush( chnl[ii].com_fd, TCIOFLUSH );
  }
  term_clear( 0 );
  term_video_off();
  exit(0);
}  /* terminate */

/*****************************************************************************

	EQN_LOOP

	Set internal loop back for "eqn" channels.  Note that only data
	is looped back, not control signals.

*****************************************************************************/

int eqn_loop( int fd, int mode, pid_t drvr_pid )
{
union async_msg msg;
struct stat xstat;

  if ( fstat( fd, &xstat) == -1 )
    return 2;
  memset( &msg, 0, sizeof(msg) );
  if ( mode )
    msg.chstatus.type = EQNCHLOOPON;
  else
    msg.chstatus.type = EQNCHLOOPOFF;
  msg.chstatus.major = major(xstat.st_rdev); 
  msg.chstatus.minor = minor(xstat.st_rdev) - 1;
  if ( Send( drvr_pid, &msg, &msg, sizeof(msg), sizeof(msg) ) )
    return 1;  /* an error */
  else
    return 0;
}  /* eqn_loop */

/*****************************************************************************

	EQNBENCH_USAGE

*****************************************************************************/

void eqnbench_usage()
{
  fprintf( stderr, "\neqnbench usage:\n\n" );
  fprintf( stderr, "  eqnbench [-l] [-v] [-f] [-b baudrate] [-c count] <port name>\n\n" );
  fprintf( stderr, "  where -l enables data loopback         (default:off)\n" );
  fprintf( stderr, "        -v enables data verification     (default:off)\n" );
  fprintf( stderr, "        -f disables isflow/osflow        (default:on)\n" );
  fprintf( stderr, "        -b indicates the baudrate to use (default:115200)\n" );
  fprintf( stderr, "        -c indicates the count of ports  (default:1)\n" );
  fprintf( stderr, "           (must be in same connector group)\n\n" );

  fprintf( stderr, "	examples:	eqnbench -l -v -c16 /dev/ttb1\n" );
  fprintf( stderr, "			eqnbench /dev/a/tta1\n\n" );
}  /* eqnbench_usage */
 
